#/usr/bin/env bash

# default value for the run name
run_name="latest"

port="8080"
allowed_host=""

zdb_container="zdb:1.2"
dir=$(pwd)

# those parameters are mostly here for debugging purposes
zdb_folder="/home/chlamdb/"
metagenlab_folder=""
debug=false
use_dev_server=false
singularity_dir="$(pwd)/singularity"


for i in "$@"; do
	case $i in
		--port=*)
			port="${i#*=}"
			shift
			;;
		--allowed_host=*)
			allowed_host="${i#*=}"
			shift
			;;
		--name=*)
			run_name="${i#*=}"
			shift
			;;
		--singularity_dir=*)
			singularity_dir="${i#*=}"
			shift
			;;
		-d|--debug)
			debug=true
			shift
			;;
		--dev_server)
			use_dev_server=true
			shift
			;;
		--zdb_folder=*)
			zdb_folder="${i#*=}"
			use_custom_zdb=true
			shift
			;;
		--metagenlab_folder=*)
			metagenlab_folder="${i#*=}"
			shift
			;;
		--dir=*)
			dir="${i#*=}"
			shift
			;;
		-h|--help)
			echo ""
			echo "This script starts the web application, using by default the database "
			echo "that was generated by the latest run of analysis"
			echo ""
			echo "Arguments:"
			echo "--dir: the directory where the results (should contain a zdb subfolder) can be found"
			echo "       (defaults to the current directory)"
			echo "--name=NAME: tells the web application to use a different database than the latest"
			echo "--port=PORT: the web server will listen on a different port (default: 8080)"
			echo "--allowed_host=HOSTS: coma separated list that will be passed as argument to django ALLOWED_HOSTS. If none specified, will try to guess with the hostname command. This is basically the URL or IP adress you will using to access the web page."
			echo "--singularity_dir: the directory where the singularity images are downloaded (default singularity in current directory)"
			echo ""
			exit 0
			;;
		*)
			echo "Unknown option $i"
			exit 1
	esac
done

debugging_mode=""


bind_path=""
if [ "$use_custom_zdb" = true ]; then
	# bind the source code of an alternative, it will be executed in the
	# zdb container instead of the "official" code stored in the container
	# Useful for testing new features and debugging
	zdb_folder=$(realpath $zdb_folder)
	bind_path="$zdb_folder"
fi

bind_path="$bind_path,${dir}/zdb/assets/:${zdb_folder}/assets/temp/"
bind_path="$bind_path,${dir}/zdb/gunicorn:/usr/local/gunicorn/"
bind_path="$bind_path,${dir}/zdb/nginx:/usr/local/nginx"

if [ "$debug" = true ]; then
	debugging_mode="-d"
fi

# De-alias the run_name
if [ -f "${dir}/zdb/results/.completed_runs/${run_name}" ]; then
	run_name=$(cat $dir/zdb/results/.completed_runs/${run_name})
else
	echo "No run with name : ${run_name} in ${dir}/zdb/results/"
	exit 1
fi

if [ ! -d "${dir}/zdb/nginx/var" ]; then
	mkdir ${dir}/zdb/nginx/var
fi

if [ ! -d "${dir}/zdb/assets" ]; then
	mkdir ${dir}/zdb/assets
fi

if [ ! -z "${metagenlab_folder}" ]; then
	bind_path="${bind_path},${metagenlab_folder}"
	export PYTHONPATH="$metagenlab_folder"
fi

if [ ! -z "${allowed_host}" ]; then
	allowed_host="--hosts=${allowed_host}"
fi

dev_server=""
if [ "$use_dev_server" = true ]; then
	dev_server="--use_dev_server"
	debugging_mode="-d"
fi


# NOTE: as some of these files may be symbolic links (particularly if using latest)
# zdb relies on this script being run from the nextflow directory so that singularity
# mounts the whole results dir in the container and keep the symbolic links valid

bind_path="${bind_path},${dir}/zdb/results/db/:${zdb_folder}/assets/db/"
bind_path="${bind_path},${dir}/zdb/results/search_index/:${zdb_folder}/assets/search_index/"
bind_path="${bind_path},${dir}/zdb/results/blast_DB/:${zdb_folder}/assets/blast_DB/"
bind_path="${bind_path},${dir}/zdb/results/alignments/${run_name}:${zdb_folder}/assets/alignments/"


# annoying part to avoid error messages from nginx attempting to write
# into an hard-coded /var/log/nginx/error.log -> the zdb/nginx/var directory is mounted
# on /var/log/nginx in the container so that nginx can write in the file and be happy.

bind_path="${bind_path},${dir}/zdb/nginx/var:/var/log/nginx"

# To be added later, to allow users to download the newick trees
# bind_path="${bind_path},zdb/results/gene_phylogenies/${run_name}:${zdb_folder}/assets/phylogenies/"

if [ ! -d "${singularity_dir}" ]; then
	mkdir -p ${singularity_dir}
fi

if [ ! -f "${singularity_dir}/${zdb_container}.sif" ]; then
	echo "Preparing the zdb container, this may take a while."
	singularity build ${singularity_dir}/${zdb_container}.sif docker://registry.hub.docker.com/metagenlab/${zdb_container} &> singularity_log
	if [ ! $? -eq 0 ]; then
		echo "Failed to download singularity, more information in the singularity_log file"
		exit 1
	else
		rm singularity_log
	fi
	echo "Done"
fi

singularity run --bind ${bind_path} \
	${singularity_dir}/${zdb_container}.sif ${zdb_folder}/start_webapp \
	--run_name=${run_name} --port=${port} ${debugging_mode} ${allowed_host} ${dev_server}
